#
# pem interface,
# the code is based on the code for the pgp interface

#
proc Pem_Help {  } {
    Help PEM    ;# Make Key button is embedded in the HTML
}

#
proc Pem_HelpOld {  } {
    global exmh

    set label "Help on setting up and using PEM"
    if [Exwin_Toplevel .pemhelp "PEM Help" PemHelp] {
	Widget_Label .pemhelp.but label {left fill} -text $label
	Widget_AddBut .pemhelp.but setup "Make Key..." [list Pem_Setup]
    
	set t [Widget_Text .pemhelp 30 -setgrid true]
	Ftoc_ColorConfigure $t
	$t insert insert "EXMH Version: $exmh(version)\n\n"
	if [catch {open "$exmh(library)/help.PEM" r} in] {
	    $t insert insert "Cannot find file help.PEM to display"
	    $t configure -state disabled
	} else {
	    $t insert insert [read $in]
	}
    }
}

#
proc Pem_ExmhExtractKeys { } {
    global msg

    set file $msg(path)
    set pemfile [Mime_TempFile "decrypt"]

    PemExec [list $file -o $pemfile -u yes] message

    exec rm -f $pemfile

    return 1
}

# remove any pem/pgp-action header and put in a brand new one
proc Pem_SeditEncrypt { action draft t } {
    global pem pemMiscRE

    SeditSave $draft $t

    set hack_id [SeditId $draft]

    set linenb 1
    set line [$t get $linenb.0 $linenb.end]

    while {![regexp $pemMiscRE(headerend) $line]} {
	if [regexp -nocase {^p(em|gp)-action:} $line] {
	    set line " dummy"
	    while {[regexp "^\[ \t]" $line]} {
		$t delete $linenb.0 [expr {$linenb + 1}].0
		set line [$t get $linenb.0 $linenb.end]
	    }
	} else {
	    set linenb [expr {$linenb + 1}]
	}
	set line [$t get $linenb.0 $linenb.end]
    }
    if {! $pem(enabled)} {
	SeditMsg $t "PEM Not enabled"
	return
    }
    set pemaction "Pem-Action: $action"

    if [regexp {encrypt} $action] {
	set recipients [join [PemMatch_Whom $draft] ",\n\t\t    "]
	catch {
	    append pemaction ";\n\trecipients=\"$recipients\""
	}
    }
    $t insert 1.0 "$pemaction\n"
}

#
proc Pem_Process { srcfile dstfile } {
    global env pem pemMiscRE

    set orig [open $srcfile r]

    # get the header of the draft and split it into mime and non-mime headers
    set mailheaders [Pgp_Misc_Segregate line \
	    {[regexp $pemMiscRE(mimeheaders) $line]} [Pgp_Misc_GetHeader $orig]]
    set mimeheaders [lindex $mailheaders 0]
    set mailheaders [lindex $mailheaders 1]

    # add the default pem-action header
    lappend mailheaders "pem-action: none"

    if {[lsearch -glob $mimeheaders "content-type:*"] < 0} {
	lappend mimeheaders "content-type: text/plain; charset=us-ascii"
    }

    # get the pem-action header and parse it
    set pemaction [split [lindex $mailheaders \
			      [lsearch -glob $mailheaders "pem-action:*"]] ";"]
    set action [string trim [string tolower \
				 [lindex [split [lindex $pemaction 0] ":"] 1]]]
    set actionparams [lrange $pemaction 1 end]

    # if there is nothing to do, stop here
    if {"$action" == "none"} {
	close $orig
	error "no action"
    }

    # parse pem-action parameters
    foreach param $actionparams {
	if [regexp {^([^=]+)=(.*)$} $param {} name value] {
	    set actionparam([string tolower [string trim $name]]) \
		[string trim $value " \t\""]
	} else {
	    error "<PEM> incorrect syntax: $param"
	}
    }

    # setup default params
    set typeparams "; x-action=$action"
    
    # get the ids of the recipients (if necessary)
    if [regexp {encrypt} $action] {
	if [info exists actionparam(recipients)] {
	    set ids $actionparam(recipients)
	} else {
	    set ids [PemMatch_Whom $srcfile]
	}

	regsub -all "\t" $ids {} ids
	regsub -all "\n" $ids {} ids
	regsub -all ",( )*" $ids {, } ids
	append typeparams ";\n\tx-recipients=\"$ids\""
    }
    
    # remove pem-action and mime-version headers
    set mailheaders [Pgp_Misc_Filter line {![regexp "^(mime-version|pem-action):" \
					     $line]} $mailheaders]

    set pemheaders $mimeheaders

    # write the message to be enrypted
    set msgfile [Mime_TempFile "msg"]
    set msg [open $msgfile w 0600]
    foreach line $pemheaders { puts $msg $line }
    puts $msg ""
    puts -nonewline $msg [read $orig]
    close $orig
    close $msg

    set pemfile [Mime_TempFile "pemfile"]

    if [catch {
	switch -exact $action {
	    encrypt	{ PemExec_Encrypt $msgfile $pemfile $ids }
	    signclear	{ PemExec_Sign $msgfile $pemfile }
	    default	{ error "<PEM> unknown action $action" }
	}
    } err] {
	exec rm -f $msgfile
	error $err
    }

    # complete mailheaders with the application/pem content-type
    if {[info exists actionparam(localaction)] && \
	    [regexp -nocase $pemMiscRE(true) $actionparam(localaction)]} { 
	set mailheaders {}
    }
    lappend mailheaders \
	    "mime-version: 1.0" \
	    "content-type: application/pem; format=mime$typeparams" \
	    "content-transfer-encoding: 7bit"

    # write out the new mail file
    set dst [open $dstfile w 0600]
    foreach line $mailheaders { puts $dst $line }
    puts $dst ""
    set msg [open $pemfile r]
    set newmsg [read $msg]
    regsub -all "\x0d" $newmsg "" newmsg
    puts -nonewline $dst $newmsg
    close $msg
    close $dst

    exec rm -f $msgfile $pemfile
}

# returns the list of recipients of the draft
proc PemMatch_Whom { draft } {
    global pem

    set ids {}

    # get the list of recipients with "whom" (does alias expansion)
    catch {exec whom -nocheck $draft} recipients
    foreach id [split $recipients "\n"] {
	if [regexp "^ *-" $id] {
	    continue
	}
	set id [string trim $id]
	regsub { at } $id {@} id
	set ids [concat $ids $id]
    }

    return $ids
}

#
proc Pem_ShowMessage { tkw part } {
    global mime mimeHdr pem pemMiscRE

    set in [open $mimeHdr($part,file) r]
    gets $in firstLine
    close $in

    # let's get the format
    if {![info exists mimeHdr($part,param,format)]} {
	lappend mimeHdr($part,params) format
	set mimeHdr($part,param,format) text
    }
    set format $mimeHdr($part,param,format)

    # the action pem performed
    if {![info exists mimeHdr($part,param,x-action)]} {
	set in [open $mimeHdr($part,file) r]
	gets $in firstLine

	gets $in line
	while {![regexp $pemMiscRE(headerend) $line]} {
	    if [regexp -nocase {^proc-type:} $line] {
		if [regexp -nocase {.*MIC-CLEAR.*} $line] {
		    set action signclear
		} elseif [regexp -nocase {.*ENCRYPTED.*} $line] {
		    set action encrypt
		}
	    }
	    gets $in line
	}
	close $in
	
    } else {
	set action $mimeHdr($part,param,x-action)
    }

    # short cut if you don't have PEM at all
    if {!$pem(enabled) && ("$action" != "signclear")} {
	Mime_ShowDefault $tkw $part
	return
    }

    # see if we should decode the thing
    if {![info exists mimeHdr($part,pemdecode)]} {
	set mimeHdr($part,pemdecode) [expr {$pem(enabled) && \
					[expr $pem(decode,$pem(showpemparts))]}]
	if $pem(enabled) {
	    MimeMenuAdd $part checkbutton \
		    -label "$pem(menutext,$action) with PEM..." \
		    -command [list busy MimeRedisplayPart $tkw $part] \
		    -variable mimeHdr($part,pemdecode)
	}
    }
    
    if {($format == "mime") || ($format == "text")} {
	if $mimeHdr($part,pemdecode) {
	    set tmpfile [Mime_TempFile "decrypt"]
	    
	    PemExec_Decrypt $mimeHdr($part,file) $tmpfile msg

	    set msg [string trim $msg " \t\n"]
	    if {![file exists $tmpfile]} {
		set msg "PEM failed to decrypt.\n$msg"
	    }

	    if {($msg != {}) && ($msg != "child process exited abnormally")} {
#farger p signaturer starter helt og holdent her
 		#set curBgCol [$tkw cget -background]
		set newBgCol "light coral"
		if [regexp "Good" $msg] {
		    set newBgCol "medium spring green"
		}
   		$tkw tag configure [MimeLabel $part part] -background \
   		    $newBgCol \
   		    -foreground [$tkw cget -foreground]
#og slutter helt og holdent her
		$tkw insert insert "$msg\n"
		MimeInsertSeparator $tkw $part 5
	    }

	    if [catch {open $tmpfile r} fileIO] {
		return
	    }
	    exec rm -f $tmpfile

	    if {$format == "mime"} {
		if {![info exists mimeHdr($part,numParts)]} {
		    set mimeHdr($part,numParts) \
			[MimeParseSingle $tkw $part $fileIO]
		    set mimeHdr($part=1,color) \
			[MimeDarkerColor $tkw $mimeHdr($part,color)]
		}
		MimeShowPart $tkw $part=1 [MimeLabel $part part] 1
	    } else {
		$tkw insert insert "[read $fileIO]"
	    }
	    MimeClose $fileIO
        } else {
	    if {$action == "signclear"} {
		$tkw insert insert "The signature hasn't been checked "
		if $pem(enabled) {
		    TextButton $tkw "Check Signature" \
			[list $mimeHdr($part,menu) invoke \
			     "$pem(menutext,$action) with PEM..."]
		}
		$tkw insert insert "\n"
		MimeInsertSeparator $tkw $part 5
		
		if [catch {Pem_Unsign [Pgp_Misc_FileString \
					   $mimeHdr($part,file)]} msg] {
		    $tkw insert insert "  Can't find the signed message.\nPlease check it out: it might be suspicious !\n"
		    return
		}
		if {$format == "mime"} {
		    set tmpfile "$mimeHdr($part,file).msg"
		    Pgp_Misc_StringFile $msg $tmpfile
		    set fileIO [open $tmpfile r]
		    exec rm -f $tmpfile
		    if {![info exists mimeHdr($part,numParts)]} {
			set mimeHdr($part,numParts) \
			    [MimeParseSingle $tkw $part $fileIO]
			set mimeHdr($part=1,color) \
			    [MimeDarkerColor $tkw $mimeHdr($part,color)]
		    }
		    MimeShowPart $tkw $part=1 [MimeLabel $part part] 1
		    MimeClose $fileIO
		} else {
		    $tkw insert insert $msg
		}
	    } else {
		Mime_ShowDefault $tkw $part
	    }
        }
    } else {
	$tkw insert insert "PEM application format '$format' unknown\n"
	return
    }
}

# unsign a pem signed message (take the pem stuff out)
proc Pem_Unsign { text } {
    if {![regexp "^(.*\n)?-+BEGIN PRIVACY-ENH\[^\n]*\n(.*)\n-+END PRIVACY-ENH" \
	      $text {} {} text]} {
	error "<Pgp_Misc_Unsign> can't find the message"
    }
    regsub "(\[^\n]+\n)*\n" $text {} text
    return "$text"
}

#
proc PemExec { arglist outvar } {
    upvar $outvar output
    global env pem pemPass

    if {!$pem(enabled)} {
	error "<PEM> pem isn't enabled"
    }
    
    if {!$pem(keeppin)} {
	Pem_ClearPassword
    }
    
    if {($pem(keeppin)) && ($pemPass == {})} {
	Pem_GetPass
	if {($pemPass == {})} {
	    #cancel button pressed
	    set output {Cancel button pressed.}
	    return 1
	}
    }

    if {$pem(crlcheck)} {
	set arglist [concat [list -R -i] $arglist]
    } else {
	set arglist [concat [list -i] $arglist]
    }


    if {!$pem(keeppin) || ($pemPass == {})} {
	return [PemExec_Interactive $arglist output]
    } else {
	return [PemExec_Batch pem $arglist output $pemPass]
    }
}

#
proc PemExec_Batch { command arglist outvar {passwd {}} } {
    upvar $outvar output
    global env pem

    set tclcmd [concat \
 	    [list exec $pem(path)/secude $command ] \
	    $arglist]

    if {($passwd == {})} {
	catch { unset env(USERPIN) }
    } else {
	set env(USERPIN) $passwd
    }

    set logfile [Mime_TempFile "pem"]

    set tclcmd [concat $tclcmd [list {>&} $logfile]]

    set result [catch $tclcmd]

    catch { unset env(USERPIN) }

    if [catch {open $logfile r} log] {
	set output {}
    } else {
	set output [read $log]
	close $log
    }

    exec rm -f $logfile

    # clean up the output
    regsub -all "\[\x0d\x07]" $output {} output
    regsub ".*SECUDE.*(poses.)" $output {} output
    regsub "Fingerprint of PK is .*" $output {} output
    regsub -all "MIC Validated.*" $output {} output
    regsub -all "0x0" $output {} output

    regsub -all "(Message signed by .?purported.? <.*\n)" \
	$output "BAD SIGNATURE\n\\1" output
    if ![regexp "BAD" $output] {
	regsub -all "(Message signed by .*\n)" \
	    $output "Good Signature\n\\1" output
    }

    set output [string trim $output]
    return $result
}

#
proc PemExec_Interactive { arglist outvar } {
    upvar $outvar output
    global pem

    set shcmd "
        $pem(path)/secude pem \"[join [Pgp_Misc_Map x {
	    regsub {([$"\`])} $x {\\1} x
	    set dummy $x
        } $arglist] {" "}]\";
	echo
	echo press Return...;
        read dummy"

    set logfile [Mime_TempFile "xterm"]
    set tclcmd {exec xterm -l -lf $logfile -title PEM -e sh -c $shcmd}

    set result [catch $tclcmd]
    if [catch {open $logfile r} log] {
	set output {}
    } else {
	set output [read $log]
	close $log
    }

    # clean up the output
    regsub -all "\[\x0d\x07]" $output {} output
    regsub ".*SECUDE.*(Enter)" $output "\\1" output
    regsub -all "Enter PIN for .*(Message signed)" \
           $output "\\1" output
    regsub "Fingerprint of PK is .*" $output {} output
    regsub -all "MIC Validated.*" $output {} output
    #regsub "\\(null\\)" $output {} output
    #regsub -all "(\n\n)*" $output {} output
    regsub -all "0x0" $output {} output

    regsub -all "(Message signed by .?purported.? <.*\n)" \
	$output "BAD SIGNATURE\n\\1" output
    if ![regexp "BAD" $output] {
	regsub -all "(Message signed by .*\n)" \
	    $output "Good Signature\n\\1" output
    }

    set output [string trim $output]

    if {$output == {}} {
      set output {No status returned from xterm.}
    }

    return $result
}

#
proc PemExec_Init {  } {
    global env pem pemPass

    if {[info exists pem(path)] && \
	    ([string length [string trim $pem(path)]] > 0) && \
	    ([lsearch -exact [split $env(PATH) :] $pem(path)] < 0)} {
	set env(PATH) $env(PATH):$pem(path)
    }

    set pemPass {}
}

#
proc PemExec_Encrypt { in out tokeys } {
    global pem pemPass hackVar

    regsub -all ",( )*" $tokeys { } tokeys

    if {($pem(keeppin))} {
	if {($pemPass == {})} {
	    Pem_GetPass
	    if {($pemPass == {})} {
		error "Cancel button pressed"
	    }
	}

	if {$pem($hackVar,fpath)} {
	    if {$pem(crlcheck)} {
		PemExec_Batch pem \
		    [concat encrypted -R -i $in -o $out -j -r $tokeys -C] \
		    output $pemPass
	    } else {
		PemExec_Batch pem \
		    [concat encrypted -i $in -o $out -j -r $tokeys -C] \
		    output $pemPass
	    }
	} else {
	    if {$pem(crlcheck)} {
		PemExec_Batch pem \
		    [concat encrypted -R -i $in -o $out -j -r $tokeys -C cert] \
		    output $pemPass
	    } else {
		PemExec_Batch pem \
		    [concat encrypted -i $in -o $out -j -r $tokeys -C cert] \
		    output $pemPass
	    }
	}
    } else {
	if {$pem($hackVar,fpath)} {
	    if {$pem(crlcheck)} {
		PemExec_Interactive \
		    [concat encrypted -R -i $in -o $out -j -r $tokeys -C] \
		    output
	    } else {
		PemExec_Interactive \
		    [concat encrypted -i $in -o $out -j -r $tokeys -C] \
		    output
	    }
	} else {
	    if {$pem(crlcheck)} {
		PemExec_Interactive \
		    [concat encrypted -R -i $in -o $out -j -r $tokeys -C cert] output
	    } else {
		PemExec_Interactive \
		    [concat encrypted -i $in -o $out -j -r $tokeys -C cert] output
	    }
	}
    }

    if {![file exists $out]} {
	error "PEM refused to generate the encrypted text.\n\nNo PIN code?"
    } else {
	return {}
    }
}

#
proc PemExec_Sign { in out } {
    global pem pemPass hackVar

    if {($pem(keeppin))} {
	if {($pemPass == {})} {
	    Pem_GetPass
	    if {($pemPass == {})} {
		error "Cancel button pressed"
	    }
	}

	if {$pem($hackVar,fpath)} {
	    if {$pem(crlcheck)} {
		PemExec_Batch pem \
		    [list mic-clear -R -i $in -o $out -C] output $pemPass
	    } else {
		PemExec_Batch pem \
		    [list mic-clear -i $in -o $out -C] output $pemPass
	    }
	} else {
	    if {$pem(crlcheck)} {
		PemExec_Batch pem \
		    [list mic-clear -R -i $in -o $out -C cert] output $pemPass
	    } else {
		PemExec_Batch pem \
		    [list mic-clear -i $in -o $out -C cert] output $pemPass
	    }
	}
    } else {
	if {$pem($hackVar,fpath)} {
	    if {$pem(crlcheck)} {
		PemExec_Interactive [list mic-clear -R -i $in -o $out -C] \
		    output
	    } else {
		PemExec_Interactive [list mic-clear -i $in -o $out -C] \
		    output
	    }
	} else {
	    if {$pem(crlcheck)} {
		PemExec_Interactive [list mic-clear -R -i $in -o $out -C cert] output
	    } else {
		PemExec_Interactive [list mic-clear -i $in -o $out -C cert] output
	    }
	}
    }

    if {![file exists $out]} {
	error "PEM refused to generate the signed text.\n\nNo PIN code?"
    } else {
	return {}
    }
}

#
proc PemExec_Decrypt { in out outvar } {
    upvar $outvar output
    global pem pemPass

    if {$pem(keeppin)} {
	PemExec [list $in -o $out -u no -F] output
    } else {
	PemExec [list $in -o $out -F] output
    }
}

#
proc Pem_GetPass { } {
    global pem pemPass

    while 1 {
	if [catch {Pgp_Misc_GetPass  "" "Enter PEM PIN" "PIN"} result] {
	    set pemPass {}
	    return
	}

	if [busy PemExec_CheckPassword $result] {
	    set pemPass $result
	    after [expr $pem(passtimeout) * 60 * 1000] \
		[list Pem_ClearPassword]
	    return
	}
    }
}

#
proc Pem_ClearPassword { } {
    global pemPass

    catch { unset pemPass }
    set pemPass {}
}

#
proc PemExec_CheckPassword { passwd } {
    global env

    if {($passwd == {})} {
	catch { unset env(USERPIN) }
	return 0
    }

    set tmpfile [Mime_TempFile "quit"]
    set logfile [Mime_TempFile "logf"]
 
    set tclcmd "exec secude psemaint < $tmpfile >& $logfile"

    set out [open $tmpfile w 0600]
    puts $out "quit"
    close $out

    set env(USERPIN) $passwd

    set result [catch $tclcmd]

    catch { unset env(USERPIN) }

    if [catch {open $logfile r} log] {
	  set output {}
    } else {
	set output [read $log]
	close $log
    }

    exec rm -f $tmpfile
    exec rm -f $logfile

    if {($output == {})} {
	# Probably cannot find pem to execute.
	Exmh_Status !${output}!
	error "<PEM> can't find pem"
    } elseif {[regexp "Invalid PIN" $output]} {
	ExmhLog "<Pem_GetPass> $output"
	return 0
    } elseif {[regexp "PSE" $output]} {
	return 1
    } else {
	error "Not able to check PIN code.\n\n$output"
    }
    
    return 0
}

#
proc Pem_Setup {} {
    global env pem

    Widget_Toplevel .newpem "Setup PEM environment"

    # figure out the path for pem files
    if [info exists env(PSE)] {
	set pem(psepath) $env(PSE)
    } elseif [file isfile $env(HOME)/pse] {
	set pem(psepath) $env(HOME)/pse
    } else {
	set pem(psepath) [pwd]/pse
    }

    if {![file exists $pem(psepath)]} {
	return [Pem_SetupInner]
    }

    Widget_Message .newpem msg -aspect 1000 -text "
You seem to have a Private Security Environment (PSE) generated
already in:

    $pem(psepath) 

This would be destroyed if you were to carry out this operation. 
The path searched for an existing PSE is:

    \$PSE
    \$HOME/pse
    ./pse      (maybe, depends on installation)"

    Widget_Frame .newpem rim Pad {top expand fill}
    .newpem.rim configure -bd 10
    Widget_Frame .newpem.rim but Menubar {top fill}

    Widget_AddBut .newpem.rim.but no "Cancel" { 
	destroy .newpem
	return 0 
    }

    tkwait window .newpem
}

#
proc Pem_SetupInner {} {
    global env pem

    set dninit $pem(dn)

    if {[catch {exec cat /etc/passwd | grep $env(USER)\: \
		| cut -d\: -f5 | cut -d\, -f1} username]} {
	if {[catch {exec ypcat passwd | grep $env(USER)\: \
		    | cut -d\: -f5 | cut -d\, -f1} username]} {
	    set username "My Name"
        }
    }

    Widget_Message .newpem msg -aspect 1000 -text \
"To be able to use the SECUDE system, you need a PIN-code (password).
Enter a PIN-code of your choice in both of the boxes below. Make
sure you remember this code when you want to use secure E-Mail later.

You will also have to state your Distinguished Name (DN), i.e. your 
X.400 mail address. If you are not sure that the address listed in 
the box is correct, please check it up before continuing.

After having pressed the Submit button, you will have to wait for
your Private Security Environment to be generated."

    Widget_Frame .newpem pin Pad {top expand fill}
    .newpem.pin configure -bd 4

    Widget_Label .newpem.pin l {left} -text "PIN: "
    Widget_Entry .newpem.pin e {left fill} -width 20 -show *
    Widget_Label .newpem.pin l2 {left} -text " (same) "
    Widget_Entry .newpem.pin e2 {left fill} -width 20 -show *

    Widget_Frame .newpem rim Pad {top expand fill}
    .newpem.rim configure -bd 4

    Widget_Label .newpem.rim l {left} -text "DN: "
    Widget_Entry .newpem.rim e {left fill} -width \
	[expr [string length $username]+[string length $dninit]+5]
    .newpem.rim.e insert 0 cn=$username\,$dninit

    Widget_Frame .newpem.rim but Menubar {top fill}
    Widget_AddBut .newpem.rim.but no "Cancel" { 
        destroy .newpem 
        return 0 
    }
    Widget_AddBut .newpem.rim.but yes "Submit" {
	if {[string length [.newpem.pin.e get]] \
		&& [.newpem.pin.e get] == [.newpem.pin.e2 get]} {
            .newpem.msg config -cursor watch
            .newpem.pin config -cursor watch
            .newpem.pin.l config -cursor watch
            .newpem.pin.e config -cursor watch
            .newpem.pin.l2 config -cursor watch
            .newpem.pin.e2 config -cursor watch
            .newpem.rim config -cursor watch
            .newpem.rim.l config -cursor watch
            .newpem.rim.e config -cursor watch
            .newpem.rim.but config -cursor watch
            .newpem.rim.but.no config -cursor watch
            .newpem.rim.but.yes config -cursor watch
            update idletasks
	    Pem_SetupExecute
	    return 1
	} else {
	    tkwait window .newpem
	}
    }

    update idletasks
    tkwait window .newpem
}

#
proc Pem_SetupExecute {} {
    global env pem mh_path
    set ca $pem(ca)
    set ra $pem(ra)
    set dn [.newpem.rim.e get]
    set pw [.newpem.pin.e get]

    set envSet 0

    # Finally ready to set up the stuff
    if {![info exists env(PSE)]} {
        set env(PSE) $pem(psepath)
        set envSet 1
    }

    #create a PSE
    PemExec_Batch psecrt [list $dn] output $pw

    #create prototype certificate
    PemExec_Batch pem [list cert-req -o /tmp/$env(USER).my.pto ] output $pw

    #extract PK fingerprint
    PemExec_Batch psemaint [list show Cert] output $pw

    regsub "(.*Public Key Fingerprint:   )(\[0-9,A-F, ]*)(\n.*)" $output "\\2" pkfingerp

    destroy .newpem.msg
    destroy .newpem.pin
    destroy .newpem.rim

    Widget_Message .newpem msg -aspect 1000 -text \
"A prototype of your personal certificate has been generated for 
you and is now being sent to your Certification Authority (CA):

         $ca,

with a copy to your Registration Authority (RA) who is:

         $ra.

You will now have to visit your RA to show proof of identity and 
bringing along this string: 

         $pkfingerp

and state that this is your public key fingerprint. You will not 
receive any certificate before your identity is verified by your RA.

When a certificate has been issued for you, you will receive
it by e-mail from your CA. Until so happens, you will not be able 
to send Privacy Enhanced Mail.

It will be necessary to restart Exmh after this key generation
operation, in order to bring forward all menues that have to do 
with PEM. You should further read the Help page where you 
pressed the Make-Key button."

    Widget_Frame .newpem but Pad {top expand fill}
    .newpem.but configure -bd 10

    Widget_AddBut .newpem.but ok "Dismiss" {destroy .newpem}

    update idletasks

    #submit prototype to CA
    exec $mh_path/mhmail $ca -cc $ra -subject "Certification request" \
                         < /tmp/$env(USER).my.pto

    #clean up
    exec rm /tmp/$env(USER).my.pto

    if {$envSet} {
      catch { unset env(PSE) }
    }

    tkwait window .newpem
}
